#ifdef _WIN32
#include <windows.h>
#else // _WIN32
#include "../winlnxdefs.h"
#endif // _WIN32
#include <GL/gl.h>
#include "glext.h"
#include "glide.h"
#include "main.h"

static int fct[4], source0[4], operand0[4], source1[4], operand1[4], source2[4], operand2[4];
static int fcta[4],sourcea0[4],operanda0[4],sourcea1[4],operanda1[4],sourcea2[4],operanda2[4];
static int alpha_ref, alpha_func;

float texture_env_color[4];
int fog_enabled;

int need_lambda[2];
float lambda_color[2][4];

void updateCombiner(int i)
{
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
	glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB, fct[i]);
	glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB, source0[i]);
	glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, operand0[i]);
	glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB, source1[i]);
	glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB, operand1[i]);
	glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE2_RGB_ARB, source2[i]);
	glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND2_RGB_ARB, operand2[i]);
}

void updateCombinera(int i)
{
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
	glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_ALPHA_ARB, fcta[i]);
	glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_ALPHA_ARB, sourcea0[i]);
	glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_ALPHA_ARB, operanda0[i]);
	glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_ALPHA_ARB, sourcea1[i]);
	glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_ALPHA_ARB, operanda1[i]);
	glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE2_ALPHA_ARB, sourcea2[i]);
	glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND2_ALPHA_ARB, operanda2[i]);
}

void init_combiner()
{
	int texture[4] = {0, 0, 0, 0};

	glActiveTextureARB(GL_TEXTURE0_ARB);
	glEnable(GL_TEXTURE_2D);

	// creating a fake texture
	glBindTexture(GL_TEXTURE_2D, 32*1024*1024);
	glTexImage2D(GL_TEXTURE_2D, 0, 3, 2, 2, 0, GL_RGBA, GL_UNSIGNED_BYTE, texture);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

	glActiveTextureARB(GL_TEXTURE1_ARB);
	glBindTexture(GL_TEXTURE_2D, 32*1024*1024);
	glEnable(GL_TEXTURE_2D);
	if (nbTextureUnits > 2)
	{
		glActiveTextureARB(GL_TEXTURE2_ARB);
		glBindTexture(GL_TEXTURE_2D, 32*1024*1024);
		glEnable(GL_TEXTURE_2D);
	}
	if (nbTextureUnits > 3)
	{
		glActiveTextureARB(GL_TEXTURE3_ARB);
		glBindTexture(GL_TEXTURE_2D, 32*1024*1024);
		glEnable(GL_TEXTURE_2D);
	}
	alpha_ref = 0;
	fog_enabled = 0;
	need_lambda[0] = need_lambda[1] = 0;
}

FX_ENTRY void FX_CALL 
grConstantColorValue( GrColor_t value )
{
	texture_env_color[3] = ((value >> 24) & 0xFF) / 255.0f;
	texture_env_color[0] = ((value >> 16) & 0xFF) / 255.0f;
	texture_env_color[1] = ((value >>  8) & 0xFF) / 255.0f;
	texture_env_color[2] = (value & 0xFF) / 255.0f;
	
	if (!need_lambda[0])
	{
		glActiveTextureARB(GL_TEXTURE0_ARB);
		glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, texture_env_color);
	}
	if (!need_lambda[1])
	{
		glActiveTextureARB(GL_TEXTURE1_ARB);
		glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, texture_env_color);
	}
	if (nbTextureUnits > 2)
	{
		glActiveTextureARB(GL_TEXTURE2_ARB);
		glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, texture_env_color);
	}
	if (nbTextureUnits > 3)
	{
		glActiveTextureARB(GL_TEXTURE3_ARB);
		glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, texture_env_color);
	}
}

int setOtherColorSource(int other)
{
	switch(other)
	{
	case GR_COMBINE_OTHER_ITERATED:
		return GL_PRIMARY_COLOR_ARB;
		break;
	case GR_COMBINE_OTHER_TEXTURE:
		return GL_PREVIOUS_ARB;
		break;
	case GR_COMBINE_OTHER_CONSTANT:
		return GL_CONSTANT_ARB;
		break;
	default:
		display_warning((unsigned char*)"unknwown other color source : %x", other);
	}
	return 0;
}

int setLocalColorSource(int local)
{
	switch(local)
	{
	case GR_COMBINE_LOCAL_ITERATED:
		return GL_PRIMARY_COLOR_ARB;
		break;
	case GR_COMBINE_LOCAL_CONSTANT:
		return GL_CONSTANT_ARB;
		break;
	default:
		display_warning((unsigned char*)"unknwown local color source : %x", local);
	}
	return 0;
}

FX_ENTRY void FX_CALL 
grColorCombine(
               GrCombineFunction_t function, GrCombineFactor_t factor,
               GrCombineLocal_t local, GrCombineOther_t other,
               FxBool invert )
{
	int num_tex;
	if (invert) display_warning((unsigned char*)"grColorCombine : inverted result");
	if (nbTextureUnits > 2)
	{
		glActiveTextureARB(GL_TEXTURE2_ARB);
		num_tex = 2;
	}
	else 
	{
		glActiveTextureARB(GL_TEXTURE1_ARB);
		num_tex = 1;
	}

	fct[num_tex+1] = GL_REPLACE;
	source0[num_tex+1] = GL_PREVIOUS_ARB;
	operand0[num_tex+1] = GL_SRC_COLOR;
	
	switch(function)
	{
	case GR_COMBINE_FUNCTION_LOCAL:
		fct[num_tex] = GL_REPLACE;
		source0[num_tex] = setLocalColorSource(local);
		operand0[num_tex] = GL_SRC_COLOR;
		break;
	case GR_COMBINE_FUNCTION_SCALE_OTHER:
		switch(factor)
		{
		case GR_COMBINE_FACTOR_LOCAL:
			fct[num_tex] = GL_MODULATE;
			source0[num_tex] = setLocalColorSource(local);
			operand0[num_tex] = GL_SRC_COLOR;
			source1[num_tex] = setOtherColorSource(other);
			operand1[num_tex] = GL_SRC_COLOR;
			break;
		case GR_COMBINE_FACTOR_ONE:
			fct[num_tex] = GL_REPLACE;
			source0[num_tex] = setOtherColorSource(other);
			operand0[num_tex] = GL_SRC_COLOR;
			break;
		case GR_COMBINE_FACTOR_ONE_MINUS_LOCAL:
			fct[num_tex] = GL_MODULATE;
			source0[num_tex] = setLocalColorSource(local);
			operand0[num_tex] = GL_ONE_MINUS_SRC_COLOR;
			source1[num_tex] = setOtherColorSource(other);
			operand1[num_tex] = GL_SRC_COLOR;
			break;
		case GR_COMBINE_FACTOR_OTHER_ALPHA:
			fct[num_tex] = GL_MODULATE;
			source0[num_tex] = setOtherColorSource(other);
			operand0[num_tex] = GL_SRC_ALPHA;
			source1[num_tex] = setOtherColorSource(other);
			operand1[num_tex] = GL_SRC_COLOR;
			break;
		case GR_COMBINE_FACTOR_TEXTURE_RGB:
			fct[num_tex] = GL_MODULATE;
			source0[num_tex] = GL_PREVIOUS_ARB;
			operand0[num_tex] = GL_SRC_COLOR;
			source1[num_tex] = setOtherColorSource(other);
			operand1[num_tex] = GL_SRC_COLOR;
			break;
		default:
			display_warning((unsigned char*)"grColorCombine(GR_COMBINE_FUNCTION_SCALE_OTHER) : unknown factor : %x", factor);
		}
		break;
	case GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL:
		switch(factor)
		{
		case GR_COMBINE_FACTOR_TEXTURE_RGB:
			fct[num_tex] = GL_MODULATE;
			source0[num_tex] = GL_PREVIOUS_ARB;
			operand0[num_tex] = GL_SRC_COLOR;
			source1[num_tex] = setOtherColorSource(other);
			operand1[num_tex] = GL_SRC_COLOR;
			fct[num_tex+1] = GL_ADD;
			source0[num_tex+1] = GL_PREVIOUS_ARB;
			operand0[num_tex+1] = GL_SRC_COLOR;
			source1[num_tex+1] = setLocalColorSource(local);
			operand1[num_tex+1] = GL_SRC_COLOR;
			if (source1[num_tex+1] == GL_PREVIOUS_ARB) source1[num_tex+1] = GL_TEXTURE0_ARB;
			break;
		case GR_COMBINE_FACTOR_ONE:
			fct[num_tex] = GL_ADD;
			source0[num_tex] = setOtherColorSource(other);
			operand0[num_tex] = GL_SRC_COLOR;
			source1[num_tex] = setLocalColorSource(local);
			operand1[num_tex] = GL_SRC_COLOR;
			break;
		case GR_COMBINE_FACTOR_LOCAL:
			fct[num_tex] = GL_MODULATE;
			source0[num_tex] = setLocalColorSource(local);
			operand0[num_tex] = GL_SRC_COLOR;
			source1[num_tex] = setOtherColorSource(other);
			operand1[num_tex] = GL_SRC_COLOR;
			fct[num_tex+1] = GL_ADD;
			source0[num_tex+1] = GL_PREVIOUS_ARB;
			operand0[num_tex+1] = GL_SRC_COLOR;
			source1[num_tex+1] = setLocalColorSource(local);
			operand1[num_tex+1] = GL_SRC_COLOR;
			if (source1[num_tex+1] == GL_PREVIOUS_ARB) source1[num_tex+1] = GL_TEXTURE0_ARB;
			break;
		case GR_COMBINE_FACTOR_LOCAL_ALPHA:
			fct[num_tex] = GL_MODULATE;
			source0[num_tex] = setLocalColorSource(local);
			operand0[num_tex] = GL_SRC_ALPHA;
			source1[num_tex] = setOtherColorSource(other);
			operand1[num_tex] = GL_SRC_COLOR;
			fct[num_tex+1] = GL_ADD;
			source0[num_tex+1] = GL_PREVIOUS_ARB;
			operand0[num_tex+1] = GL_SRC_COLOR;
			source1[num_tex+1] = setLocalColorSource(local);
			operand1[num_tex+1] = GL_SRC_COLOR;
			if (source1[num_tex+1] == GL_PREVIOUS_ARB) source1[num_tex+1] = GL_TEXTURE0_ARB;
			break;
		case GR_COMBINE_FACTOR_ONE_MINUS_LOCAL:
			fct[num_tex] = GL_MODULATE;
			source0[num_tex] = setLocalColorSource(local);
			operand0[num_tex] = GL_ONE_MINUS_SRC_COLOR;
			source1[num_tex] = setOtherColorSource(other);
			operand1[num_tex] = GL_SRC_COLOR;
			fct[num_tex+1] = GL_ADD;
			source0[num_tex+1] = GL_PREVIOUS_ARB;
			operand0[num_tex+1] = GL_SRC_COLOR;
			source1[num_tex+1] = setLocalColorSource(local);
			operand1[num_tex+1] = GL_SRC_COLOR;
			if (source1[num_tex+1] == GL_PREVIOUS_ARB) source1[num_tex+1] = GL_TEXTURE0_ARB;
			break;
		case GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA:
			fct[num_tex] = GL_MODULATE;
			source0[num_tex] = setLocalColorSource(local);
			operand0[num_tex] = GL_ONE_MINUS_SRC_ALPHA;
			source1[num_tex] = setOtherColorSource(other);
			operand1[num_tex] = GL_SRC_COLOR;
			fct[num_tex+1] = GL_ADD;
			source0[num_tex+1] = GL_PREVIOUS_ARB;
			operand0[num_tex+1] = GL_SRC_COLOR;
			source1[num_tex+1] = setLocalColorSource(local);
			operand1[num_tex+1] = GL_SRC_COLOR;
			if (source1[num_tex+1] == GL_PREVIOUS_ARB) source1[num_tex+1] = GL_TEXTURE0_ARB;
			break;
		default:
			display_warning((unsigned char*)"grColorCombine(GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL) : unknown factor : %x", factor);
		}
		break;
	case GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL:
		switch(factor)
		{
		case GR_COMBINE_FACTOR_TEXTURE_RGB:
			// this combiner mode isn't implemented correctly and doesn't work on a TNT
			fct[num_tex] = GL_SUBTRACT_ARB;
			source0[num_tex] = setOtherColorSource(other);
			operand0[num_tex] = GL_SRC_COLOR;
			source1[num_tex] = setLocalColorSource(local);
			operand1[num_tex] = GL_SRC_COLOR;
			fct[num_tex+1] = GL_MODULATE;
			source0[num_tex+1] = GL_TEXTURE0_ARB;
			operand0[num_tex+1] = GL_SRC_COLOR;
			source1[num_tex+1] = GL_PREVIOUS_ARB;
			operand1[num_tex+1] = GL_SRC_COLOR;
			break;
		case GR_COMBINE_FACTOR_ONE:
			// doesn't work on TNT
			fct[num_tex] = GL_SUBTRACT_ARB;
			source0[num_tex] = setOtherColorSource(other);
			operand0[num_tex] = GL_SRC_COLOR;
			source1[num_tex] = setLocalColorSource(local);
			operand1[num_tex] = GL_SRC_COLOR;
			break;
		default:
			display_warning((unsigned char*)"grColorCombine(GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL) : unknown factor : %x", factor);
		}
		break;
	case GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL:
		switch(factor)
		{
		case GR_COMBINE_FACTOR_OTHER_ALPHA:
			fct[num_tex] = GL_INTERPOLATE_ARB;
			source0[num_tex] = setOtherColorSource(other);
			operand0[num_tex] = GL_SRC_COLOR;
			source1[num_tex] = setLocalColorSource(local);
			operand1[num_tex] = GL_SRC_COLOR;
			source2[num_tex] = setOtherColorSource(other);
			operand2[num_tex] = GL_SRC_ALPHA;
			break;
		case GR_COMBINE_FACTOR_LOCAL_ALPHA:
			fct[num_tex] = GL_INTERPOLATE_ARB;
			source0[num_tex] = setOtherColorSource(other);
			operand0[num_tex] = GL_SRC_COLOR;
			source1[num_tex] = setLocalColorSource(local);
			operand1[num_tex] = GL_SRC_COLOR;
			source2[num_tex] = setLocalColorSource(local);
			operand2[num_tex] = GL_SRC_ALPHA;
			break;
		case GR_COMBINE_FACTOR_TEXTURE_ALPHA:
			fct[num_tex] = GL_INTERPOLATE_ARB;
			source0[num_tex] = setOtherColorSource(other);
			operand0[num_tex] = GL_SRC_COLOR;
			source1[num_tex] = setLocalColorSource(local);
			operand1[num_tex] = GL_SRC_COLOR;
			source2[num_tex] = GL_PREVIOUS_ARB;
			operand2[num_tex] = GL_SRC_ALPHA;
			break;
		case GR_COMBINE_FACTOR_TEXTURE_RGB:
			fct[num_tex] = GL_INTERPOLATE_ARB;
			source0[num_tex] = setOtherColorSource(other);
			operand0[num_tex] = GL_SRC_COLOR;
			source1[num_tex] = setLocalColorSource(local);
			operand1[num_tex] = GL_SRC_COLOR;
			source2[num_tex] = GL_PREVIOUS_ARB;
			operand2[num_tex] = GL_SRC_COLOR;
			break;
		default:
			display_warning((unsigned char*)"grColorCombine(GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL) : unknown factor : %x", factor);
		}
		break;
	case GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL:
		switch(factor)
		{
		case GR_COMBINE_FACTOR_TEXTURE_RGB:
			fct[num_tex] = GL_MODULATE;
			source0[num_tex] = GL_PREVIOUS_ARB;
			operand0[num_tex] = GL_ONE_MINUS_SRC_COLOR;
			source1[num_tex] = setLocalColorSource(local);
			operand1[num_tex] = GL_SRC_COLOR;
			break;
		default:
			display_warning((unsigned char*)"grColorCombine(GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL) : unknown factor : %x", factor);
		}
		break;
	default:
		display_warning((unsigned char*)"grColorCombine : unknown function : %x", function);
	}
	updateCombiner(num_tex);
	if (nbTextureUnits > 3)
	{
		glActiveTextureARB(GL_TEXTURE3_ARB);
		updateCombiner(num_tex+1);
	}
}

int setOtherAlphaSource(int other)
{
	switch(other)
	{
	case GR_COMBINE_OTHER_ITERATED:
		return GL_PRIMARY_COLOR_ARB;
		break;
	case GR_COMBINE_OTHER_TEXTURE:
		return GL_PREVIOUS_ARB;
		break;
	case GR_COMBINE_OTHER_CONSTANT:
		return GL_CONSTANT_ARB;
		break;
	default:
		display_warning((unsigned char*)"unknwown other alpha source : %x", other);
	}
	return 0;
}

int setLocalAlphaSource(int local)
{
	switch(local)
	{
	case GR_COMBINE_LOCAL_ITERATED:
		return GL_PRIMARY_COLOR_ARB;
		break;
	case GR_COMBINE_LOCAL_CONSTANT:
		return GL_CONSTANT_ARB;
		break;
	default:
		display_warning((unsigned char*)"unknwown local alpha source : %x", local);
	}
	return 0;
}

FX_ENTRY void FX_CALL
grAlphaCombine(
               GrCombineFunction_t function, GrCombineFactor_t factor,
               GrCombineLocal_t local, GrCombineOther_t other,
               FxBool invert
               )
{
	int num_tex;
	if (invert) display_warning((unsigned char*)"grAlphaCombine : inverted result");
	if (nbTextureUnits > 2)
	{
		glActiveTextureARB(GL_TEXTURE2_ARB);
		num_tex = 2;
	}
	else 
	{
		glActiveTextureARB(GL_TEXTURE1_ARB);
		num_tex = 1;
	}

	fcta[num_tex+1] = GL_REPLACE;
	sourcea0[num_tex+1] = GL_PREVIOUS_ARB;
	operanda0[num_tex+1] = GL_SRC_ALPHA;

	switch(function)
	{
	case GR_COMBINE_FUNCTION_LOCAL:
		fcta[num_tex] = GL_REPLACE;
		sourcea0[num_tex] = setLocalAlphaSource(local);
		operanda0[num_tex] = GL_SRC_ALPHA;
		break;
	case GR_COMBINE_FUNCTION_SCALE_OTHER:
		switch (factor)
		{
		case GR_COMBINE_FACTOR_ONE:
			fcta[num_tex] = GL_REPLACE;
			sourcea0[num_tex] = setOtherAlphaSource(other);
			operanda0[num_tex] = GL_SRC_ALPHA;
			break;
		case GR_COMBINE_FACTOR_LOCAL:
			fcta[num_tex] = GL_MODULATE;
			sourcea0[num_tex] = setLocalAlphaSource(local);
			operanda0[num_tex] = GL_SRC_ALPHA;
			sourcea1[num_tex] = setOtherAlphaSource(other);
			operanda1[num_tex] = GL_SRC_ALPHA;
			break;
		case GR_COMBINE_FACTOR_ONE_MINUS_LOCAL:
			fcta[num_tex] = GL_MODULATE;
			sourcea0[num_tex] = setLocalAlphaSource(local);
			operanda0[num_tex] = GL_ONE_MINUS_SRC_ALPHA;
			sourcea1[num_tex] = setOtherAlphaSource(other);
			operanda1[num_tex] = GL_SRC_ALPHA;
			break;
		case GR_COMBINE_FACTOR_TEXTURE_ALPHA:
			fcta[num_tex] = GL_MODULATE;
			sourcea0[num_tex] = GL_PREVIOUS_ARB;
			operanda0[num_tex] = GL_SRC_ALPHA;
			sourcea1[num_tex] = setOtherAlphaSource(other);
			operanda1[num_tex] = GL_SRC_ALPHA;
			break;
		case GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA:
			fcta[num_tex] = GL_MODULATE;
			sourcea0[num_tex] = GL_PREVIOUS_ARB;
			operanda0[num_tex] = GL_ONE_MINUS_SRC_ALPHA;
			sourcea1[num_tex] = setOtherAlphaSource(other);
			operanda1[num_tex] = GL_SRC_ALPHA;
			break;
		default:
			display_warning((unsigned char*)"grAlphaCombine(GR_COMBINE_FUNCTION_SCALE_OTHER) : unknown factor : %x", factor);
		}
		break;
	case GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL:
		switch(factor)
		{
		case GR_COMBINE_FACTOR_TEXTURE_ALPHA:
			fcta[num_tex] = GL_MODULATE;
			sourcea0[num_tex] = setOtherAlphaSource(other);
			operanda0[num_tex] = GL_SRC_ALPHA;
			sourcea1[num_tex] = GL_PREVIOUS_ARB;
			operanda0[num_tex] = GL_SRC_ALPHA;
			fcta[num_tex+1] = GL_ADD;
			sourcea0[num_tex+1] = GL_PREVIOUS_ARB;
			operanda0[num_tex+1] = GL_SRC_ALPHA;
			sourcea1[num_tex+1] = setLocalAlphaSource(local);
			operanda1[num_tex+1] = GL_SRC_ALPHA;
			if (sourcea1[num_tex+1] == GL_PREVIOUS_ARB) sourcea1[num_tex+1] = GL_TEXTURE0_ARB;
			break;
		case GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA:
			fcta[num_tex] = GL_MODULATE;
			sourcea0[num_tex] = setOtherAlphaSource(other);
			operanda0[num_tex] = GL_SRC_ALPHA;
			sourcea1[num_tex] = GL_PREVIOUS_ARB;
			operanda0[num_tex] = GL_ONE_MINUS_SRC_ALPHA;
			fcta[num_tex+1] = GL_ADD;
			sourcea0[num_tex+1] = GL_PREVIOUS_ARB;
			operanda0[num_tex+1] = GL_SRC_ALPHA;
			sourcea1[num_tex+1] = setLocalAlphaSource(local);
			operanda1[num_tex+1] = GL_SRC_ALPHA;
			if (sourcea1[num_tex+1] == GL_PREVIOUS_ARB) sourcea1[num_tex+1] = GL_TEXTURE0_ARB;
			break;
		case GR_COMBINE_FACTOR_LOCAL:
			fcta[num_tex] = GL_MODULATE;
			sourcea0[num_tex] = setOtherAlphaSource(other);
			operanda0[num_tex] = GL_SRC_ALPHA;
			sourcea1[num_tex] = setLocalAlphaSource(local);
			operanda0[num_tex] = GL_SRC_ALPHA;
			fcta[num_tex+1] = GL_ADD;
			sourcea0[num_tex+1] = GL_PREVIOUS_ARB;
			operanda0[num_tex+1] = GL_SRC_ALPHA;
			sourcea1[num_tex+1] = setLocalAlphaSource(local);
			operanda1[num_tex+1] = GL_SRC_ALPHA;
			if (sourcea1[num_tex+1] == GL_PREVIOUS_ARB) sourcea1[num_tex+1] = GL_TEXTURE0_ARB;
			break;
		case GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA:
			fcta[num_tex] = GL_MODULATE;
			sourcea0[num_tex] = setOtherAlphaSource(other);
			operanda0[num_tex] = GL_SRC_ALPHA;
			sourcea1[num_tex] = setLocalAlphaSource(local);
			operanda0[num_tex] = GL_ONE_MINUS_SRC_ALPHA;
			fcta[num_tex+1] = GL_ADD;
			sourcea0[num_tex+1] = GL_PREVIOUS_ARB;
			operanda0[num_tex+1] = GL_SRC_ALPHA;
			sourcea1[num_tex+1] = setLocalAlphaSource(local);
			operanda1[num_tex+1] = GL_SRC_ALPHA;
			if (sourcea1[num_tex+1] == GL_PREVIOUS_ARB) sourcea1[num_tex+1] = GL_TEXTURE0_ARB;
			break;
		case GR_COMBINE_FACTOR_ONE:
			fcta[num_tex] = GL_ADD;
			sourcea0[num_tex] = setOtherAlphaSource(other);
			operanda0[num_tex] = GL_SRC_ALPHA;
			sourcea1[num_tex] = setLocalAlphaSource(local);
			operanda1[num_tex] = GL_SRC_ALPHA;
			break;
		default:
			display_warning((unsigned char*)"grAlphaCombine(GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL) : unknown factor : %x", factor);
		}
		break;
	case GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL:
		switch(factor)
		{
		case GR_COMBINE_FACTOR_TEXTURE_ALPHA:
			// this combiner mode isn't implemented correctly and doesn't work on a TNT
			fcta[num_tex] = GL_SUBTRACT_ARB;
			sourcea0[num_tex] = setOtherAlphaSource(other);
			operanda0[num_tex] = GL_SRC_ALPHA;
			sourcea1[num_tex] = setLocalAlphaSource(local);
			operanda1[num_tex] = GL_SRC_ALPHA;
			fcta[num_tex+1] = GL_MODULATE;
			sourcea0[num_tex+1] = GL_TEXTURE0_ARB;
			operanda0[num_tex+1] = GL_SRC_ALPHA;
			sourcea1[num_tex+1] = GL_PREVIOUS_ARB;
			operanda1[num_tex+1] = GL_SRC_ALPHA;
			break;
		case GR_COMBINE_FACTOR_LOCAL:
			fcta[num_tex] = GL_SUBTRACT_ARB;
			sourcea0[num_tex] = setOtherAlphaSource(other);
			operanda0[num_tex] = GL_SRC_ALPHA;
			sourcea1[num_tex] = setLocalAlphaSource(local);
			operanda1[num_tex] = GL_SRC_ALPHA;
			fcta[num_tex+1] = GL_MODULATE;
			sourcea0[num_tex+1] = setLocalAlphaSource(local);
			operanda0[num_tex+1] = GL_SRC_ALPHA;
			sourcea1[num_tex+1] = GL_PREVIOUS_ARB;
			operanda1[num_tex+1] = GL_SRC_ALPHA;
			break;
		case GR_COMBINE_FACTOR_ONE:
			// doesn't work on a TNT
			fcta[num_tex] = GL_SUBTRACT_ARB;
			sourcea0[num_tex] = setOtherAlphaSource(other);
			operanda0[num_tex] = GL_SRC_ALPHA;
			sourcea1[num_tex] = setLocalAlphaSource(local);
			operanda1[num_tex] = GL_SRC_ALPHA;
			break;
		default:
			display_warning((unsigned char*)"grAlphaCombine(GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL) : unknown factor : %x", factor);
		}
		break;
	case GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL:
		switch(factor)
		{
		case GR_COMBINE_FACTOR_TEXTURE_ALPHA:
			fcta[num_tex] = GL_INTERPOLATE_ARB;
			sourcea0[num_tex] = setOtherAlphaSource(other);
			operanda0[num_tex] = GL_SRC_ALPHA;
			sourcea1[num_tex] = setLocalAlphaSource(local);
			operanda1[num_tex] = GL_SRC_ALPHA;
			sourcea2[num_tex] = GL_PREVIOUS_ARB;
			operanda2[num_tex] = GL_SRC_ALPHA;
			break;
		default:
			display_warning((unsigned char*)"grAlphaCombine(GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL) : unknown factor : %x", factor);
		}
		break;
	case GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL:
		switch(factor)
		{
		case GR_COMBINE_FACTOR_TEXTURE_ALPHA:
			fcta[num_tex] = GL_MODULATE;
			sourcea0[num_tex] = GL_PREVIOUS_ARB;
			operanda0[num_tex] = GL_ONE_MINUS_SRC_ALPHA;
			sourcea1[num_tex] = setLocalAlphaSource(local);
			operanda1[num_tex] = GL_SRC_ALPHA;
			break;
		default:
			display_warning((unsigned char*)"grAlphaCombine(GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL) : unknown factor : %x", factor);
		}
		break;
	default:
		display_warning((unsigned char*)"grAlphaCombine : unknown function : %x", function);
	}
	updateCombinera(num_tex);
	if (nbTextureUnits > 3)
	{
		glActiveTextureARB(GL_TEXTURE3_ARB);
		updateCombinera(num_tex+1);
	}
}

FX_ENTRY void FX_CALL 
grTexCombine(
             GrChipID_t tmu,
             GrCombineFunction_t rgb_function,
             GrCombineFactor_t rgb_factor, 
             GrCombineFunction_t alpha_function,
             GrCombineFactor_t alpha_factor,
             FxBool rgb_invert,
             FxBool alpha_invert
             )
{
	int num_tex;
	int will_need_lambda = 0;
	if (rgb_invert) display_warning((unsigned char*)"grTexCombine : inverted rgb result");
	if (alpha_invert) display_warning((unsigned char*)"grTexCombine : inverted alpha result");

	if (tmu == GR_TMU1 && nbTextureUnits <= 2) return;
	if (tmu == GR_TMU0)
	{
		if (nbTextureUnits <= 2)
		{
			glActiveTextureARB(GL_TEXTURE0_ARB);
			num_tex = 0;
		}
		else
		{
			glActiveTextureARB(GL_TEXTURE1_ARB);
			num_tex = 1;
		}
	}
	else
	{
		glActiveTextureARB(GL_TEXTURE0_ARB);
		num_tex = 0;
	}

	switch(rgb_function)
	{
	case GR_COMBINE_FUNCTION_ZERO:
		fct[num_tex] = GL_REPLACE;
		source0[num_tex] = GL_PRIMARY_COLOR_ARB;
		operand0[num_tex] = GL_SRC_COLOR;
		break;
	case GR_COMBINE_FUNCTION_LOCAL:
		fct[num_tex] = GL_REPLACE;
		source0[num_tex] = GL_TEXTURE;
		operand0[num_tex] = GL_SRC_COLOR;
		break;
	case GR_COMBINE_FUNCTION_SCALE_OTHER:
		switch(rgb_factor)
		{
		case GR_COMBINE_FACTOR_LOCAL:
			fct[num_tex] = GL_MODULATE;
			source0[num_tex] = GL_PREVIOUS_ARB;
			operand0[num_tex] = GL_SRC_COLOR;
			source1[num_tex] = GL_TEXTURE;
			operand1[num_tex] = GL_SRC_COLOR;
			break;
		case GR_COMBINE_FACTOR_LOCAL_ALPHA:
			fct[num_tex] = GL_MODULATE;
			source0[num_tex] = GL_PREVIOUS_ARB;
			operand0[num_tex] = GL_SRC_COLOR;
			source1[num_tex] = GL_TEXTURE;
			operand1[num_tex] = GL_SRC_ALPHA;
			break;
		case GR_COMBINE_FACTOR_ONE:
			fct[num_tex] = GL_REPLACE;
			source0[num_tex] = GL_PREVIOUS_ARB;
			operand0[num_tex] = GL_SRC_COLOR;
			break;
		default:
			display_warning((unsigned char*)"grTextCombine(GR_COMBINE_FUNCTION_SCALE_OTHER) : unknown rgb factor : %x", rgb_factor);
		}
		break;
	case GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL:
		switch(rgb_factor)
		{
		case GR_COMBINE_FACTOR_ONE:
			fct[num_tex] = GL_ADD;
			source0[num_tex] = GL_PREVIOUS_ARB;
			operand0[num_tex] = GL_SRC_COLOR;
			source1[num_tex] = GL_TEXTURE;
			operand1[num_tex] = GL_SRC_COLOR;
			break;
		case GR_COMBINE_FACTOR_DETAIL_FACTOR:
			if (num_tex==1)
			{
				fct[num_tex-1] = GL_MODULATE;
				source0[num_tex-1] = GL_TEXTURE;
				operand0[num_tex-1] = GL_SRC_COLOR;
				source1[num_tex-1] = GL_CONSTANT_ARB;
				operand1[num_tex-1] = GL_SRC_ALPHA;
				fct[num_tex] = GL_ADD;
				source0[num_tex] = GL_PREVIOUS_ARB;
				operand0[num_tex] = GL_SRC_COLOR;
				source1[num_tex] = GL_TEXTURE;
				operand1[num_tex] = GL_SRC_COLOR;
				will_need_lambda = 1;
			}
			break;
		default:
			display_warning((unsigned char*)"grTextCombine(GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL) : unknown rgb factor : %x", rgb_factor);
		}
		break;
	case GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL:
		switch(rgb_factor)
		{
		case GR_COMBINE_FACTOR_DETAIL_FACTOR:
			fct[num_tex] = GL_INTERPOLATE_ARB;
			source0[num_tex] = GL_PREVIOUS_ARB;
			operand0[num_tex] = GL_SRC_COLOR;
			source1[num_tex] = GL_TEXTURE;
			operand1[num_tex] = GL_SRC_COLOR;
			source2[num_tex] = GL_CONSTANT_ARB;
			operand2[num_tex] = GL_SRC_ALPHA;
			will_need_lambda = 1;
			break;
		default:
			display_warning((unsigned char*)"grTextCombine(GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL) : unknown rgb factor : %x", rgb_factor);
		}
		break;
	case GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL:
		switch(rgb_factor)
		{
		case GR_COMBINE_FACTOR_LOCAL:
			// this combiner mode isn't implemented correctly and doesn't work on a TNT
			if (num_tex==1)
			{
				fct[num_tex-1] = GL_SUBTRACT_ARB;
				source0[num_tex-1] = GL_TEXTURE0_ARB;
				operand0[num_tex-1] = GL_SRC_COLOR;
				source1[num_tex-1] = GL_TEXTURE1_ARB;
				operand1[num_tex-1] = GL_SRC_COLOR;
				fct[num_tex] = GL_MODULATE;
				source0[num_tex] = GL_PREVIOUS_ARB;
				operand0[num_tex] = GL_SRC_COLOR;
				source1[num_tex] = GL_TEXTURE;
				operand1[num_tex] = GL_SRC_COLOR;
			}
			break;
		default:
			display_warning((unsigned char*)"grTextCombine(GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL) : unknown rgb factor : %x", rgb_factor);
		}
	default:
		display_warning((unsigned char*)"grTextCombine : unknown rgb function : %x", rgb_function);
	}

	switch(alpha_function)
	{
	case GR_COMBINE_FUNCTION_ZERO:
		fcta[num_tex] = GL_REPLACE;
		sourcea0[num_tex] = GL_PRIMARY_COLOR_ARB;
		operanda0[num_tex] = GL_SRC_ALPHA;
		break;
	case GR_COMBINE_FUNCTION_LOCAL:
		fcta[num_tex] = GL_REPLACE;
		sourcea0[num_tex] = GL_TEXTURE;
		operanda0[num_tex] = GL_SRC_ALPHA;
		break;
	case GR_COMBINE_FUNCTION_SCALE_OTHER:
		switch(alpha_factor)
		{
		case GR_COMBINE_FACTOR_LOCAL:
			fcta[num_tex] = GL_MODULATE;
			sourcea0[num_tex] = GL_TEXTURE;
			operanda0[num_tex] = GL_SRC_ALPHA;
			sourcea1[num_tex] = GL_PREVIOUS_ARB;
			operanda1[num_tex] = GL_SRC_ALPHA;
			break;
		case GR_COMBINE_FACTOR_ONE:
			fcta[num_tex] = GL_REPLACE;
			sourcea0[num_tex] = GL_PREVIOUS_ARB;
			operanda0[num_tex] = GL_SRC_ALPHA;
			break;
		default:
			display_warning((unsigned char*)"grTextCombine(GR_COMBINE_FUNCTION_SCALE_OTHER) : unknown alpha factor : %x", alpha_factor);
		}
		break;
	case GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL:
		switch(alpha_factor)
		{
		case GR_COMBINE_FACTOR_ONE:
			fcta[num_tex] = GL_ADD;
			sourcea0[num_tex] = GL_PREVIOUS_ARB;
			operanda0[num_tex] = GL_SRC_ALPHA;
			sourcea1[num_tex] = GL_TEXTURE;
			operanda1[num_tex] = GL_SRC_ALPHA;
			break;
		case GR_COMBINE_FACTOR_DETAIL_FACTOR:
			if (num_tex==1)
			{
				fct[num_tex-1] = GL_MODULATE;
				source0[num_tex-1] = GL_TEXTURE;
				operand0[num_tex-1] = GL_SRC_ALPHA;
				source1[num_tex-1] = GL_CONSTANT_ARB;
				operand1[num_tex-1] = GL_SRC_ALPHA;
				fct[num_tex] = GL_ADD;
				source0[num_tex] = GL_PREVIOUS_ARB;
				operand0[num_tex] = GL_SRC_ALPHA;
				source1[num_tex] = GL_TEXTURE;
				operand1[num_tex] = GL_SRC_ALPHA;
				will_need_lambda = 1;
			}
			break;
		default:
			display_warning((unsigned char*)"grTextCombine(GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL) : unknown alpha factor : %x", alpha_factor);
		}
		break;
	case GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL:
		switch(alpha_factor)
		{
		case GR_COMBINE_FACTOR_DETAIL_FACTOR:
			fcta[num_tex] = GL_INTERPOLATE_ARB;
			sourcea0[num_tex] = GL_PREVIOUS_ARB;
			operanda0[num_tex] = GL_SRC_ALPHA;
			sourcea1[num_tex] = GL_TEXTURE;
			operanda1[num_tex] = GL_SRC_ALPHA;
			sourcea2[num_tex] = GL_CONSTANT_ARB;
			operanda2[num_tex] = GL_SRC_ALPHA;
			will_need_lambda = 1;
			break;
		default:
			display_warning((unsigned char*)"grTextCombine(GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL) : unknown alpha factor : %x", alpha_factor);
		}
		break;
	case GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL:
		switch(alpha_factor)
		{
		case GR_COMBINE_FACTOR_DETAIL_FACTOR:
			fcta[num_tex] = GL_MODULATE;
			sourcea0[num_tex] = GL_CONSTANT_ARB;
			operanda0[num_tex] = GL_ONE_MINUS_SRC_ALPHA;
			sourcea1[num_tex] = GL_TEXTURE;
			operanda1[num_tex] = GL_SRC_ALPHA;
			will_need_lambda = 1;
			break;
		default:
			display_warning((unsigned char*)"grTextCombine(GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL) : unknown alpha factor : %x", alpha_factor);
		}
		break;
	case GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA:
		switch(alpha_factor)
		{
		case GR_COMBINE_FACTOR_OTHER_ALPHA:
			fcta[num_tex] = GL_MODULATE;
			sourcea0[num_tex] = GL_PREVIOUS_ARB;
			operanda0[num_tex] = GL_ONE_MINUS_SRC_ALPHA;
			sourcea1[num_tex] = GL_TEXTURE;
			operanda1[num_tex] = GL_SRC_ALPHA;
			break;
		default:
			display_warning((unsigned char*)"grTextCombine(GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA) : unknown alpha factor : %x", alpha_factor);
		}
		break;
	default:
		display_warning((unsigned char*)"grTextCombine : unknown alpha function : %x", alpha_function);
	}

	if (will_need_lambda != need_lambda[num_tex])
	{
		need_lambda[num_tex] = will_need_lambda;
		if (need_lambda[num_tex])
		{
			int i;
			for (i=0; i<3; i++) lambda_color[num_tex][i] = texture_env_color[i];
			lambda_color[num_tex][3] = lambda;
			glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, lambda_color[num_tex]);
		}
		else
			glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, texture_env_color);
	}
	updateCombiner(num_tex);
	updateCombinera(num_tex);
}

FX_ENTRY void FX_CALL
grAlphaBlendFunction(
                     GrAlphaBlendFnc_t rgb_sf,   GrAlphaBlendFnc_t rgb_df,
                     GrAlphaBlendFnc_t alpha_sf, GrAlphaBlendFnc_t alpha_df
                     )
{
	int sfactorRGB = 0, dfactorRGB = 0, sfactorAlpha = 0, dfactorAlpha = 0;

	switch(rgb_sf)
	{
	case GR_BLEND_ZERO:
		sfactorRGB = GL_ZERO;
		break;
	case GR_BLEND_SRC_ALPHA:
		sfactorRGB = GL_SRC_ALPHA;
		break;
	case GR_BLEND_ONE:
		sfactorRGB = GL_ONE;
		break;
	case GR_BLEND_ONE_MINUS_SRC_ALPHA:
		sfactorRGB = GL_ONE_MINUS_SRC_ALPHA;
		break;
	default:
		display_warning((unsigned char*)"grAlphaBlendFunction : rgb_sf = %x", rgb_sf);
	}

	switch(rgb_df)
	{
	case GR_BLEND_ZERO:
		dfactorRGB = GL_ZERO;
		break;
	case GR_BLEND_SRC_ALPHA:
		dfactorRGB = GL_SRC_ALPHA;
		break;
	case GR_BLEND_ONE:
		dfactorRGB = GL_ONE;
		break;
	case GR_BLEND_ONE_MINUS_SRC_ALPHA:
		dfactorRGB = GL_ONE_MINUS_SRC_ALPHA;
		break;
	default:
		display_warning((unsigned char*)"grAlphaBlendFunction : rgb_df = %x", rgb_df);
	}

	switch(alpha_sf)
	{
	case GR_BLEND_ZERO:
		sfactorAlpha = GL_ZERO;
		break;
	case GR_BLEND_ONE:
		sfactorAlpha = GL_ONE;
		break;
	default:
		display_warning((unsigned char*)"grAlphaBlendFunction : alpha_sf = %x", alpha_sf);
	}

	switch(alpha_df)
	{
	case GR_BLEND_ZERO:
		dfactorAlpha = GL_ZERO;
		break;
	case GR_BLEND_ONE:
		dfactorAlpha = GL_ONE;
		break;
	default:
		display_warning((unsigned char*)"grAlphaBlendFunction : alpha_df = %x", alpha_df);
	}
	glEnable(GL_BLEND);
	if (blend_func_separate_support)
		glBlendFuncSeparateEXT(sfactorRGB, dfactorRGB, sfactorAlpha, dfactorAlpha);
	else
		glBlendFunc(sfactorRGB, dfactorRGB);
}

FX_ENTRY void FX_CALL
grAlphaTestReferenceValue( GrAlpha_t value )
{
	alpha_ref = value;
	grAlphaTestFunction(alpha_func);
}

FX_ENTRY void FX_CALL
grAlphaTestFunction( GrCmpFnc_t function )
{
	alpha_func = function;
	switch(function)
	{
	case GR_CMP_GREATER:
		glAlphaFunc(GL_GREATER, alpha_ref/255.0f);
		break;
	case GR_CMP_GEQUAL:
		glAlphaFunc(GL_GEQUAL, alpha_ref/255.0f);
		break;
	case GR_CMP_ALWAYS:
		glAlphaFunc(GL_ALWAYS, alpha_ref/255.0f);
		glDisable(GL_ALPHA_TEST);
		return;
		break;
	default:
		display_warning((unsigned char*)"grAlphaTestFunction : unknown function : %x", function);
	}
	glEnable(GL_ALPHA_TEST);
}


// fog

FX_ENTRY void FX_CALL 
grFogMode( GrFogMode_t mode )
{
	switch(mode)
	{
	case GR_FOG_DISABLE:
		glDisable(GL_FOG);
		fog_enabled = 0;
		break;
	case GR_FOG_WITH_TABLE_ON_Q:
		glEnable(GL_FOG);
		fog_enabled = 1;
		break;
	default:
		display_warning((unsigned char*)"grFogMode : unknown mode : %x", mode);
	}
}

FX_ENTRY void FX_CALL
guFogGenerateLinear(GrFog_t *fogtable,
                    float nearZ, float farZ )
{
	glFogi(GL_FOG_MODE, GL_LINEAR);
	glFogi(GL_FOG_COORDINATE_SOURCE_EXT, GL_FOG_COORDINATE_EXT);
	glFogf(GL_FOG_START, nearZ / 65528.0f);
	glFogf(GL_FOG_END, farZ / 65528.0f);
}

FX_ENTRY void FX_CALL 
grFogTable( const GrFog_t ft[] )
{
}

FX_ENTRY void FX_CALL 
grFogColorValue( GrColor_t fogcolor )
{
	float color[4];
	color[3] = ((fogcolor >> 24) & 0xFF) / 255.0f;
	color[0] = ((fogcolor >> 16) & 0xFF) / 255.0f;
	color[1] = ((fogcolor >>  8) & 0xFF) / 255.0f;
	color[2] = (fogcolor & 0xFF) / 255.0f;
	glFogfv(GL_FOG_COLOR, color); 
}

// chroma

FX_ENTRY void FX_CALL 
grChromakeyMode( GrChromakeyMode_t mode )
{
	switch(mode)
	{
	case GR_CHROMAKEY_DISABLE:
		break;
	default:
		display_warning((unsigned char*)"grChromakeyMode : unknown mode : %x", mode);
	}
}
